/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: dft.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: dft.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
/*-----------------------------------------------------------------------
 *  Discrete Fourier transform package.
 *  This package performs discrete Fourier Transforms of unequally
 *  spaced data.  Applications involving regularly spaced data should
 *  us an FFT which will generally be MUCH faster.
 *     A structure containing the sines/cosines is created once and then
 *  can be used multiple times.  
 *  Complex numbers are stored as real and imaginary parts in successive 
 *  array elements.
 *
 *     The Fourier Transforms included are:
 *     DFTr2c  - Real to complex.
 *     DFTc2r  - half plane complex to real.
 *     DFTc2c  - complex to complex .
 *  To use this package:
 *   1) Create and initialize an DFT_info using makeDFT or makeDFTramp
 *   2) Call the appropriate routine FT routine
 *   3) Free the memory used by the DFT_info using killDFT
 *---------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include "dft.h"

DFT_info* makeDFT (int dir, long num_x, float* x, long num_u, 
		  float* u)
/*----------------------------------------------------------------------- */
/*   Make an DFT_info structure and fill values.                          */
/*   This structure needs to be created for a specific Fourier transform  */
/*   although once created may be used for as many repetitions of the same*/
/*   transform as are needed.                                             */
/*   The terms "X" and "U" refer to the initial and transformed domains.  */
/*   Note: x and u arrays should be in appropriate physical units and NOT */
/*   merely indices.                                                      */
/*   Inputs:                                                              */
/*       dir   The direction of the transform: FORWARD or BACKWARD        */
/*       num_x The number of elements in the X direction                  */
/*             This is the number of complex elements if x is complex     */
/*       x     Array of coordinate values on X axis                       */
/*       num_u The number of elements in the U direction                  */
/*             This is the number of complex elements if u is complex     */
/*       u     Array of coordinate values on U axis                       */
/*   Returns:                                                             */
/*      An DFT_info structure pointer ; NULL indicates failure            */
/*----------------------------------------------------------------------- */
{
  long      i, j, indx; 
  float     arg, twopi;
  float     *cc, *ss;
  DFT_info *out = (DFT_info*)malloc(sizeof(DFT_info));
  if (out==NULL) return out;  /* check allocation */

  /* set values */
  out->num_in  =  num_x;
  out->num_out =  num_u;
  out->cos_arr = (float*) malloc (sizeof(float)*num_x*num_u);
  if (out->cos_arr==NULL) {killDFT(out); return NULL;}
  out->sin_arr = (float*) malloc (sizeof(float)*num_x*num_u);
  if (out->cos_arr==NULL) {killDFT(out); return NULL;}

  /* fill in sines and cosines */
  if (dir==FORWARD) twopi = -6.2831853072;
  else  twopi = 6.2831853072;
  cc = out->cos_arr;
  ss = out->sin_arr;
  indx = 0;
  for (j=0; j<num_u; j++) {
    for (i=0; i<num_x; i++) {
      arg = twopi * x[i] * u[j];
      cc[indx]   = cos(arg);
      ss[indx++] = sin(arg);
    } /* end loop over X */
  } /* end loop over U */

  return out;
} /* end makeDFT */

void killDFT (DFT_info* DFi) 
/*----------------------------------------------------------------------- */
/*   Delete an DFT_info structure and free memory                         */
/*   Inputs:                                                              */
/*       DFi   The structure to be deleted                                */
/*----------------------------------------------------------------------- */
{
  if (DFi==NULL) return; /* check for bad pointer */
  if (DFi->cos_arr!=NULL) free(DFi->cos_arr);
  if (DFi->sin_arr!=NULL) free(DFi->sin_arr);
  free (DFi);
} /* end */

void DFTc2c (DFT_info* DFi, float in[], float out[]) 
/*----------------------------------------------------------------------- */
/*   Perform Complex to Complex Fourier transform                         */
/*   The direction, input and output coordinate values are specified when */
/*   the DFT_info DFi was created.                                        */
/*   Inputs:                                                              */
/*       DFi   The structure defining the transform, created by makeDFT   */
/*       in    Input array of values as described to makeDFT              */
/*             complex values are represented as real, imag in successive */
/*             elements of the array                                      */
/*   Outputs:                                                             */
/*       out   Output array of values as described to MakeDFT             */
/*             complex values are represented as real, imag in successive */
/*             elements of the array                                      */
/*----------------------------------------------------------------------- */
{
  double sumr, sumi;
  long   i, j, indx, num_x, num_u;
  float  *cc, *ss;
  if (DFi==NULL) return; /* check for bad pointer */
  if (in==NULL) return;  /* check for bad pointer */
  if (out==NULL) return; /* check for bad pointer */

  /* convert to local arrays */
  cc = DFi->cos_arr;
  ss = DFi->sin_arr;

  /* do transform */
  indx = 0;
  num_x = DFi->num_in * 2;
  num_u = DFi->num_out * 2;
  for (j=0; j<num_u; j+=2) {
    sumr = 0.0; sumi = 0.0;
    for (i=0; i<num_x; i+=2) {
      sumr += in[i] * cc[indx]   - in[i+1] * ss[indx];
      sumi += in[i+1] * cc[indx] + in[i] * ss[indx];
      indx++;
    } /* end loop over X */
    out[j]   = sumr;
    out[j+1] = sumi;
  } /* end loop over U */

} /* end DFTc2c */

void DFTr2c (DFT_info* DFi, float in[], float out[]) 
/*----------------------------------------------------------------------- */
/*   Perform Real to Complex Fourier transform                            */
/*   The direction, input and output coordinate values are specified when */
/*   the DFT_info DFi was created.                                        */
/*   Inputs:                                                              */
/*       DFi   The structure defining the transform, created by makeDFT   */
/*       in    Input array of real values as described to makeDFT         */
/*   Outputs:                                                             */
/*       out   Output array of values as described to MakeDFT             */
/*             complex values are represented as real, imag in successive */
/*             elements of the array                                      */
/*----------------------------------------------------------------------- */
{
  double sumr, sumi;
  long   i, j, indx, num_x, num_u;
  float  *cc, *ss;
  if (DFi==NULL) return; /* check for bad pointer */
  if (in==NULL) return;  /* check for bad pointer */
  if (out==NULL) return; /* check for bad pointer */

  /* convert to local arrays */
  cc = DFi->cos_arr;
  ss = DFi->sin_arr;

  /* do transform */
  indx = 0;
  num_x = DFi->num_in;
  num_u = DFi->num_out * 2;
  for (j=0; j<num_u; j+=2) {
    sumr = 0.0; sumi = 0.0;
    for (i=0; i<num_x; i++) {
      sumr += in[i] * cc[indx];
      sumi += in[i] * ss[indx];
      indx++;
    } /* end loop over X */
    out[j]   = sumr;
    out[j+1] = sumi;
  } /* end loop over U */
} /* end DFTr2c */

void DFTc2r (DFT_info* DFi, float in[], float out[]) 
/*----------------------------------------------------------------------- */
/*   Perform Complex to Complex Fourier transform                         */
/*   The direction, input and output coordinate values are specified when */
/*   the DFT_info DFi was created.                                        */
/*   Inputs:                                                              */
/*       DFi   The structure defining the transform, created by makeDFT   */
/*       in    Input array of values as described to makeDFT              */
/*             complex values are represented as real, imag in successive */
/*             elements of the array                                      */
/*   Outputs:                                                             */
/*       out   Output array of real values as described to MakeDFT        */
/*             complex values are represented as real, imag in successive */
/*             elements of the array                                      */
/*----------------------------------------------------------------------- */
{
  double sumr;
  long   i, j, indx, num_x, num_u;
  float  *cc, *ss;
  if (DFi==NULL) return; /* check for bad pointer */
  if (in==NULL) return;  /* check for bad pointer */
  if (out==NULL) return; /* check for bad pointer */

  /* convert to local arrays */
  cc = DFi->cos_arr;
  ss = DFi->sin_arr;

  /* do transform */
  indx = 0;
  num_x = DFi->num_in * 2;
  num_u = DFi->num_out;
  for (j=0; j<num_u; j++) {
    sumr = 0.0;
    for (i=0; i<num_x; i+=2) {
      sumr += in[i] * cc[indx] - in[i+1] * ss[indx];
      indx++;
    } /* end loop over X */
    out[j]   = sumr;
  } /* end loop over U */
} /* end DFTc2r */

/* */
